#include "hoc.h"

#if defined(HAVE_ERRNO_H)
#include <errno.h>
#endif

#if defined(HAVE_FLOAT_H)
#include <float.h>
#endif

#if defined(HAVE_MATH_H)
#include <math.h>
#endif

#if defined(DEBUG_MACHEPS)
#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif
#endif

/* These functions are in -lm, but are missing from math.h, sigh... */

#if defined(__linux)
EXTERN fp_t trunc ARGS((fp_t));		/* all GNU/Linux */
#endif

#if defined(__osf__)
EXTERN int ilogb ARGS((fp_t));		/* Compaq/DEC OSF/1 */
#if defined(HAVE_NEAREST)
EXTERN fp_t nearest ARGS((fp_t,fp_t));	/* Compaq/DEC OSF/1 */
#endif
#endif

#define TWOTO23		FP(8388608.0)					/* 2^23 */
#define TWOTO52		FP(4503599627370496.0)				/* 2^52 */
#define TWOTO64		FP(18446744073709551616.0)			/* 2^64 */
#define TWOTO112	FP(5192296858534827628530496329220096.0)	/* 2^112 */

fp_t
Ceil(fp_t x)
{
	fp_t y;

	/* TO DO: Should Ceil(-Infinity) return a finite number or
	not? Sun's returns ceil(Infinity) as 2^63, and ceil(-Infinity)
	as -2^63, which both seem inconsistent. */

	if (IsNaN(x))
		return (x);
	else if (IsInf(x))
		return (x);

#if defined(HAVE_CEIL)
	y = ceil(x);
#else
	y = (fp_t)((LONG_LONG)(x));
	if (x > y)
		y++;
#endif
	return (y);
}

fp_t
Class(fp_t x)
{
	if (IsQNaN(x))
		return (CLASS_QNAN);
	else if (IsSNaN(x))
		return (CLASS_SNAN);
	else if (IsInf(x))
		return ((x < FP(0.0)) ? CLASS_NEGINF : CLASS_POSINF);
	else if (x <= -MinNormal())
		return (CLASS_NEGNORMAL);
	else if (x <= -MinSubnormal())
		return (CLASS_NEGSUBNORMAL);
	else if ((x == FP(0.0)) && (Copysign(FP(1.0),x) == FP(-1.0)))
		return (CLASS_NEGZERO);
	else if ((x == FP(0.0)) && (Copysign(FP(1.0),x) == FP(1.0)))
		return (CLASS_POSZERO);
	else if (x >= MinNormal())
		return (CLASS_POSNORMAL);
	else if (x >= MinSubnormal())
		return (CLASS_POSSUBNORMAL);
	else
		return (CLASS_UNKNOWN);
}

fp_t
Copysign(fp_t x, fp_t y)
{
#if defined(HAVE_COPYSIGN)
	return (copysign(x,y));
#else
	execerror("copysign(x,y) is not available on this system", (const char*)NULL);
	return (FP(0.0));			/* NOT REACHED */
#endif
}

fp_t
Exponent(fp_t x)
{	/* Return the unbiased base-2 exponent of x */
#if defined(HAVE_IEEE_754)
	FP_Parts d;
	fp_t xabs;
	int big_endian = IsBigEndian();
	int offset;

	xabs = (x < FP(0.0)) ? -x : x;

	if (!IsNaN(x) && (xabs < MinNormal())) /* protect against NaN botch! */
	{

#if defined(HAVE_FP_T_DOUBLE)
		offset = -52;
		x *= TWOTO52;
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
		if (sizeof(d) == sizeof(d.v))	/* 128-bit */
		{
		    offset = -112;
		    x *= TWOTO112;
		}
		else			/* 80-bit Intel x86 format */
		{
		    offset = -64;
		    x *= TWOTO64;
		}
#endif

#if defined(HAVE_FP_T_SINGLE)
		offset = -23;
		x *= TWOTO23;
#endif

	}
	else
		offset = 0;

	d.v = x;

#if defined(HAVE_FP_T_DOUBLE)
	return ((fp_t)((int)((d.i[1-big_endian] & UL(0x7ff00000)) >> 20) - 1023 + offset));
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
	if (sizeof(d) == sizeof(d.v))	/* 128-bit */
		return ((fp_t)((int)((d.i[3*(1-big_endian)] & UL(0x7fff0000)) >> 16)) - 16383 + offset);
	else				/* 80-bit Intel x86 format */
		return ((fp_t)((int)(d.i[2] & UL(0x7fff)) - 16383 + offset));
#endif

#if defined(HAVE_FP_T_SINGLE)
	return ((fp_t)((int)((d.i[0] & UL(0x7f800000)) >> 23) - 127 + offset));
#endif

#else
	return ((fp_t)Nint(Log2(xabs))); /* TO DO: implement efficiently for non-IEEE 754 systems */
#endif
}

fp_t
Floor(fp_t x)
{
	fp_t y;

	/* TO DO: Should Floor(-Infinity) return a finite number or
	not? Sun's returns floor(Infinity) as 2^63, and floor(-Infinity)
	as -2^63, which both seem inconsistent. */

	if (IsNaN(x))
		return (x);
	else if (IsInf(x))
		return (x);

#if defined(HAVE_FLOOR)
	y = floor(x);
#else
	y = (fp_t)((LONG_LONG)(x));
	if (x < y)
		y--;
#endif
	return (y);
}

fp_t
Fmod(fp_t x, fp_t y)
{
#if defined(HAVE_FMOD)
	return (fmod(x,y));
#else
	execerror("fmod(x,y) is not available on this system", (const char*)NULL);
	return (FP(0.0));			/* NOT REACHED */
#endif
}

fp_t
Ilogb(fp_t x)
{
#if defined(HAVE_ILOGB)
	return ((fp_t)ilogb(x));
#else
	return ((fp_t)(int)Log2(x));
#endif
}

fp_t
Infinity(void)
{
	static fp_t infinity = FP(0.0);

	if (infinity == FP(0.0))	/* normally happens only on first call */
	{
#if defined(HAVE_IEEE_754)
		infinity = store(FP(1.0)) / store(FP(0.0));
#elif defined(DBL_MAX)
		infinity = DBL_MAX;
#else
		execerror("Infinity() is not available on this system",
			  (const char*)NULL);
#endif
	}
	return (infinity);
}

fp_t
Integer(fp_t x)
{
#if 0
	return (fp_t)(long)x;
#else
	/* Since all numbers are represented as type fp_t, we can
	provided a wider range of integers than long provides: */
	if (IsNaN(x))
	{
		execerror("int(NaN) is undefined", (const char*)NULL);
		return (FP(0.0));		/* NOT REACHED */
	}
	else if (x < FP(0.0))
		return (Ceil(x));
	else /* (x >= FP(0.0)) */
		return (Floor(x));
#endif
}

int
IsBigEndian(void)
{
	FP_Parts d;
	d.v = FP(1.0);
	return ((d.i[0] != 0) ? 1 : 0);
}

fp_t
IsFinite(fp_t x)
{
	return ((fp_t)(!IsNaN(x) && !IsInf(x)));
}

fp_t
IsInf(fp_t x)
{
	/* This portable implementation that does not require IEEE 754
	   arithmetic to function properly is due to Norman
	   L. Schryer, and used in his FPTEST package.  See ``A Test
	   of a Computer's Floating-Point Arithmetic Unit'', Bell
	   Laboratories Computing Science Technical Report no. 89,
	   February 4, 1981.  Available, with the software (under
	   license), at http://www.bell-labs.com/project/fptest/ */

#if defined(HAVE_NAN_BOTCH)		/* see IsNaN() for why */
	if (IsNaN(x))
		return (FP(0.0));
#endif

	if (x < FP(0.0))
		x = -x;			/* force positive */
	if (x < FP(1.0))		/* small, so finite */
		return (FP(0.0));
	else if ((x / BASE) == x)	/* only Inf/BASE == Inf */
		return (FP(1.0));
	else				/* must be finite */
		return (FP(0.0));
}

fp_t
IsNaN(fp_t x)
{
#if defined(HAVE_NAN_BOTCH)
	/* The Portland Group compilers, pgcc and pgCC, on Intel x86 */
	/* incorrectly generate a 64-bit integer comparison, instead */
	/* of a floating-point one.  This is wrong for NaNs, so even */
	/* the store() subterfuge does not produce the right answer. */
	/* Therefore, we have to grovel around in the bits.          */
	FP_Parts d;
	int big_endian = IsBigEndian();
	d.v = x;

#if defined(HAVE_FP_T_DOUBLE)
	return ((fp_t)(((d.i[1-big_endian] & UL(0x7ff00000)) == UL(0x7ff00000)) &&
			 ((d.i[1-big_endian] & UL(~0xfff00000)) ||
			  d.i[big_endian])));
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
	if (sizeof(d) == sizeof(d.v))	/* 128-bit */
		return ((fp_t)(((d.i[3*(1-big_endian)] & UL(0x7fff0000)) == UL(0x7fff0000)) &&
			       ((d.i[3*(1-big_endian)] & UL(~0x7fff0000)) ||
				d.i[1] || d.i[2] || d.i[3*big_endian])));
	else				/* 80-bit Intel x86 format */
		return ((fp_t)(((d.i[2] & UL(0x7fff)) == UL(0x7fff)) && (d.i[1] || d.i[0])));
#endif

#if defined(HAVE_FP_T_SINGLE)
	return ((fp_t)(((d.i[0] & UL(0x7f800000)) == UL(0x7f800000)) &&
		       (d.i[0] & UL(~0xff800000))));
#endif

#else
	return ((fp_t)(store(x) != x));	/* should thwart optimizers */
#endif
}

fp_t
IsNormal(fp_t x)
{
	return (fp_t)((IsFinite(x) && !IsSubnormal(x)));
}

/***********************************************************************
The implementation of IsQNaN() is unavoidably system-dependent, and for
two of these architectures, there is only one type of NaN, so for them,
both IsSNaN() and IsQNaN() should be true!

From p. 4-59 of the Alpha AXP Architecture Manual:
	quiet NaN:	initial fraction bit of 1
	signaling NaN:	initial fraction bit of 0
------------------------------------------------------------------------
From p. 21 of SPARC Architecture Manual v8:
[same choice as Alpha]
	quiet NaN:	initial fraction bit of 1
	signaling NaN:	initial fraction bit of 0
------------------------------------------------------------------------
From p. 145 of The PowerPC Architecture:
[same as Alpha and SPARC]
	quiet NaN:	initial fraction bit of 1
	signaling NaN:	initial fraction bit of 0
------------------------------------------------------------------------
From p. 2-5 of i860 Microprocessor Programmer's Reference Manual:
[same as Alpha, SPARC, PowerPC]
	quiet NaN:	initial fraction bit of 1
	signaling NaN:	initial fraction bit of 0
------------------------------------------------------------------------
From p. 2-15 of MC68881 Floating-Point Coprocessor Users Manual:
[same as Alpha, SPARC, PowerPC, Intel i860]
	quiet NaN:	initial fraction bit of 1
	signaling NaN:	initial fraction bit of 0
------------------------------------------------------------------------
From p. 5-3 (sequential 87 of 216) Intel IA-64 Architecture Software
Developer's Manual, Rev 1.0:
[same as Alpha, SPARC, PowerPC, Intel i860, MC68881]
	quiet NaN:	initial fraction bit of 1
	signaling NaN:	initial fraction bit of 0
------------------------------------------------------------------------
From p. 8-7 of PA-RISC 2.0 Architecture:
[opposite of Alpha, SPARC, PowerPC, i860]
	quiet NaN:	initial fraction bit of 0
	signaling NaN:	initial fraction bit of 1
------------------------------------------------------------------------
From p. E-2 of MIPS RISC Architecture:
[no distinction between NaNs]
	NaN:		initial fraction bit of 0
------------------------------------------------------------------------
From Palmer and Morse The 8087 Primer and p. I-25 of iAPX 286
Programmer's Reference Manual:
[no distinction between NaNs]
	NaN:		at least one nonzero fraction bit
Generated NaNs have negative sign bit.

Experiments on Pentium III show 0/0 -> 0xfff80000_00000000, so the
leading fraction bit is the only nonzero generated.  However, this can
be explicitly negated to obtain 0x7ff80000_00000000.
------------------------------------------------------------------------
***********************************************************************/

fp_t
IsQNaN(fp_t x)
{
	if (IsNaN(x))
	{
		FP_Parts d;
		int big_endian = IsBigEndian();

		d.v = x;

#if defined(HAVE_FP_T_DOUBLE)
#if defined(__alpha)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#elif defined(__sparc)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#elif defined(__i860)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#elif defined(__ia64)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#elif defined(__hppa)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#elif defined(__mips)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#else  /* fall back to majority style, which could still be wrong */
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#endif
#endif /* defined(HAVE_FP_T_DOUBLE) */

#if defined(HAVE_FP_T_QUADRUPLE)
#if defined(__alpha)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#elif defined(__sparc)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#elif defined(__i860)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#elif defined(__ia64)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#elif defined(__hppa)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#elif defined(__mips)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#else  /* fall back to majority style, which could still be wrong */
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#endif
#endif /* defined(HAVE_FP_T_QUADRUPLE) */

#if defined(HAVE_FP_T_SINGLE)
#if defined(__alpha)
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#elif defined(__sparc)
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#elif defined(__i860)
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#elif defined(__ia64)
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#elif defined(__hppa)
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#elif defined(__mips)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#else  /* fall back to majority style, which could still be wrong */
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#endif
#endif /* defined(HAVE_FP_T_DOUBLE) */

	}
	else
		return (FP(0.0));
}

fp_t
IsSNaN(fp_t x)
{
	if (IsNaN(x))
	{
		FP_Parts d;
		int big_endian = IsBigEndian();

		d.v = x;

#if defined(HAVE_FP_T_DOUBLE)
#if defined(__alpha)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#elif defined(__sparc)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#elif defined(__i860)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#elif defined(__ia64)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#elif defined(__hppa)
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) != 0));
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#elif defined(__mips)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#else  /* fall back to majority style, which could still be wrong */
		return ((fp_t)((d.i[1 - big_endian] & UL(0x00080000)) == 0));
#endif
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
#if defined(__alpha)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#elif defined(__sparc)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#elif defined(__i860)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#elif defined(__ia64)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#elif defined(__hppa)
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) != 0));
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#elif defined(__mips)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#else  /* fall back to majority style, which could still be wrong */
		return ((fp_t)((d.i[3*(1 - big_endian)] & UL(0x00008000)) == 0));
#endif
#endif

#if defined(HAVE_FP_T_SINGLE)
#if defined(__alpha)
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#elif defined(__sparc)
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#elif defined(__i860)
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#elif defined(__ia64)
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#elif defined(__hppa)
		return ((fp_t)((d.i[0] & UL(0x00400000)) != 0));
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#elif defined(__mips)
		(void)store(d.v);	/* avoid complaints about unused d */
		return ((fp_t)(big_endian == big_endian)); /* avoids compiler warnings about unused big_endian*/
#else  /* fall back to majority style, which could still be wrong */
		return ((fp_t)((d.i[0] & UL(0x00400000)) == 0));
#endif
#endif

	}
	else
		return (FP(0.0));
}

fp_t
IsSubnormal(fp_t x)
{
	if (x == FP(0.0))
		return (FP(0.0));
	else if (IsNaN(x))
		return (FP(0.0));
	else if (IsInf(x))
		return (FP(0.0));
	else
	{
#if defined(HAVE_IEEE_754)
		return ((fp_t)(Exponent(x) < EMIN));
#else
		return (FP(0.0));
#endif
	}
}

fp_t
Ldexp(fp_t x, fp_t y)
{
#if defined(HAVE_LDEXP)
	return (ldexp(x,(int)y));
#else
	return (x * pow(BASE, (fp_t)(int)y));
#endif
}

#if defined(HAVE_FP_T_QUADRUPLE)
/* Sun Solaris <sunmath.h> contains this note: ``ANSI C Future Library
   Directions {4.13.4}: Note that frexpf, ldexpf, modff, frexpl, ldexpl
   and modfl are _not_ provided. */
fp_t
ldexpl(fp_t x, int n)
{					/* return x * 2**n */
	fp_t result;

	if (IsNaN(x))
		result = x;
	else if (IsInf(x))
		result = x;
	else
	{
		result = x * powl(FP(2.0), (fp_t)n);
		if (IsInf(result))
		{
			/* Sun documentation says that ldexp() returns
			   +/-HUGE_VAL on overflow, and sets errno to
			   ERANGE */
			result = (x < FP(0.0)) ? -MaxNormal() : MaxNormal();
			errno = ERANGE;
		}
	}
	return (result);
}
#endif /* defined(HAVE_FP_T_QUADRUPLE) */

#if defined(HAVE_FP_T_SINGLE)
/* Sun Solaris <sunmath.h> contains this note: ``ANSI C Future Library
   Directions {4.13.4}: Note that frexpf, ldexpf, modff, frexpl, ldexpl
   and modfl are _not_ provided. */
fp_t
ldexpf(fp_t x, int n)
{					/* return x * 2**n */
	fp_t result;

	if (IsNaN(x))
		result = x;
	else if (IsInf(x))
		result = x;
	else
	{
		result = x * pow(FP(2.0), (fp_t)n); /* NB: pow, not powf, for macro substitution */
		if (IsInf(result))
		{
			/* Sun documentation says that ldexp() returns
			   +/-HUGE_VAL on overflow, and sets errno to
			   ERANGE */
			result = (x < FP(0.0)) ? -MaxNormal() : MaxNormal();
			errno = ERANGE;
		}
	}
	return (result);
}
#endif /* defined(HAVE_FP_T_SINGLE) */


#if defined(DEBUG_MACHEPS)
void
Dump(const char *s, fp_t x)
{
	FP_Parts d;
	int big_endian = IsBigEndian();
	d.v = x;

#if defined(HAVE_FP_T_DOUBLE)
	(void)fprintf(stderr,"%s: %g\t0x%08x_%08x\n",
		      s, d.v, d.i[1-big_endian], d.i[big_endian]);
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
	if (big_endian)
		(void)fprintf(stderr,"%s: %g\t0x%08x_%08x_%08x_%08x\n",
			      s, d.v, d.i[0], d.i[1], d.i[2], d.i[3]);
	else if (sizeof(d) == sizeof(d.v))	/* 128-bit little endian */
		(void)fprintf(stderr,"%s: %g\t0x%08x_%08x_%08x_%08x\n",
			      s, d.v, d.i[3], d.i[2], d.i[1], d.i[0]);
	else				/* 80-bit Intel x86 format */
		(void)fprintf(stderr,"%s: %g\t0x%04x_%08x_%08x\n",
			      s, d.v, (d.i[2] & 0xffff), d.i[1], d.i[0]);
#endif

#if defined(HAVE_FP_T_SINGLE)
	(void)fprintf(stderr,"%s: %g\t0x%08x\n", s, d.v, d.i[0]);
#endif

}
#endif


fp_t
Macheps(fp_t x)
{
	/* Macheps(x) returns the generalized machine epsilon of x,
	   the smallest number which, when added to x, produces a sum
	   that still differs from x: (x + macheps(x)) != x.

	   NB: This code guarantees these results:
		  macheps(Inf)  -> NaN
		  macheps(-Inf) -> NaN
		  macheps(NaN)  -> NaN
		  macheps(-NaN) -> -NaN
	*/

	fp_t epsilon;
	fp_t temp;

#if !defined(HAVE_NAN_BOTCH)
	static fp_t macheps_1 = FP(0.0);

	if ((x == FP(1.0)) && (macheps_1 != FP(0.0))) /* handle common case fast */
		return (macheps_1);
#endif

	if (IsNaN(x))
		return (x);

	if (IsInf(x))
		return (NaN());

	if (x == FP(0.0))
		epsilon = FP(1.0);
	else
		epsilon = fabs(x);

#if defined(DEBUG_MACHEPS)
	Dump("macheps(): 1: x = ", x);
	Dump("macheps(): 1: epsilon = ", epsilon);
	Dump("macheps(): 1: epsilon/BASE = ", epsilon/BASE);
#if 1
	temp = store(x + store(epsilon/BASE));
#else
	temp = store(x + epsilon/BASE);
#endif
	Dump("macheps(): 1: temp = ", temp);
	while (store(temp) != x)
	{
		epsilon /= BASE;
		Dump("macheps(): 2: epsilon = ", epsilon);
#if 1
		temp = store(x + store(epsilon/BASE));
#else
		temp = store(x + epsilon/BASE);
#endif
		Dump("macheps(): 2: temp = ", temp);
	}
#else
	while (store(x + store(epsilon/BASE)) != x)
		epsilon /= BASE;
#endif

	/* purify epsilon to eliminate all but high-order bit */
	temp = store(x + epsilon);
	if (IsInf(temp))/* x == MAXNORMAL: avoid overflow in purification */
		epsilon = store(x  + store(epsilon - x));
	else
		epsilon = store(temp - x);

#if !defined(HAVE_NAN_BOTCH)
	if (x == FP(1.0))			/* remember common case */
		macheps_1 = epsilon;
#endif

	return (epsilon);
}

fp_t
MaxNormal(void)
{
	static fp_t maxnormal = FP(0.0);

	if (maxnormal == FP(0.0))	      /* happens only on first call */
		maxnormal = (BASE * ((FP(1.0) - Pow(BASE, -P)) * Pow(BASE, EMAX)));
	return (maxnormal);
}

fp_t
MaxSubnormal(void)
{
	static fp_t maxsubnormal = FP(0.0);

	if (maxsubnormal == FP(0.0))	/* happens only on first call */
	{
		maxsubnormal = (FP(1.0) - Pow(BASE, -(P - FP(1.0)))) * MinNormal();
		if (maxsubnormal == FP(0.0)) /* then this system lacks subnormals */
			maxsubnormal = Macheps(FP(0.0));
	}
	return (maxsubnormal);
}

fp_t
MinNormal(void)
{
	static fp_t minnormal = FP(0.0);

	if (minnormal == FP(0.0))	      /* happens only on first call */
		minnormal = Pow(BASE, EMIN);
	return (minnormal);
}

fp_t
MinSubnormal(void)
{
	/* Return the minimimum subnormal number.  If subnormals are
	   not supported, then return the minimim normal number instead. */
	static fp_t minsubnormal = FP(0.0);

	if (minsubnormal == FP(0.0))	/* normally happens only on first call */
	{
		/* This code works for IEEE 754 (with or without subnormals),
		   and also non-IEEE 754. */
		minsubnormal = Macheps(FP(0.0));
	}
	return (minsubnormal);
}

fp_t
NaN(void)
{
	static fp_t nan = FP(0.0);

	if (nan == FP(0.0))	/* normally happens only on first call */
	{
#if defined(HAVE_IEEE_754)
		nan = store(FP(0.0)) / store(FP(0.0));
#else
		/* ISO/IEC 9899:1999(E) Programming languages -- C, Second
		edition 1999-12-01, says in section 7.12.11.2 on p. 236: ``If
		the implementation does not support quiet NaNs, the functions
		[nan(""), nanf(""), nanl("")] return zero.'' */
#endif
	}
	return (nan);
}

fp_t
Nearest(fp_t x, fp_t y)
{
#if defined(HAVE_NEAREST)
	return (nearest(x,y));
#else
	return (Nextafter(x,y));		/* TO BE FIXED: incorrect translation */
#endif
}

fp_t
Nextafter(fp_t x, fp_t y)
{
#if defined(HAVE_NEXTAFTER)
	return (nextafter(x,y));
#else
	if (IsNaN(x))
		return (NaN());
	else if (IsNaN(y))
		return (NaN());
	else if (IsInf(x) && (x > FP(0.0)) && ((y < FP(0.0)) || (y < x)))
		return (MaxNormal());
	else if (IsInf(x) && (x < FP(0.0)) && ((y > FP(0.0)) || (y > x)))
		return (-MaxNormal());
	else if (IsInf(x) && (x == y))
		return (x);
	else if ((x == MaxNormal()) && (x < y))
		return (Infinity());
	else if ((x == -MaxNormal()) && (x > y))
		return (-Infinity());
	else if (x < y)
		return (x + macheps(x));
	else if (x > y)
		return (x - macheps(x)/BASE);
	else /* (x == y) */
		return (x);
#endif
}

fp_t
Nint(fp_t x)
{
	return ((x >= FP(0.0)) ?
		(fp_t)((LONG_LONG)(x + FP(0.5))) :
		(fp_t)((LONG_LONG)(x - FP(0.5))));
}

fp_t
QNaN(void)
{
	return (NaN());
}

fp_t
Remainder(fp_t x, fp_t y)
{
#if defined(HAVE_REMAINDER)
	return (remainder(x,y));
#else
	execerror("remainder(x) is not available on this system", (const char*)NULL);
	return (FP(0.0));			/* NOT REACHED */
#endif
}

#if defined(HAVE_FP_T_QUADRUPLE)
fp_t
remainderl(fp_t x, fp_t y)
{
	fp_t result;

	if (y == FP(0.0))
		result = QNaN();
	else if (IsNaN(x))
		result = x;
	else if (IsNaN(y))
		result = y;
	else if (IsInf(x))
		result = QNaN();
	else if (IsInf(y))
		result = x;
	else
	{
		/* TO DO: What should happen when x/y is not representable as an integer? */
		LONG_LONG n;

		n = (LONG_LONG)Nint(x/y);
		if (fabs((fp_t)n - x/y) == FP(0.5)) /* then must ensure that n is even */
		{
			if (n < 0)
				n = (n - 1) & ~1;
			else
				n = (n + 1) & ~1;
		}
		result = x - (fp_t)n * y;
	}
	return (result);
}
#endif

fp_t
Rint(fp_t x)
{
#if defined(HAVE_RINT)
	return (rint(x));
#else
	execerror("rint(x) is not available on this system", (const char*)NULL);
	return (FP(0.0));			/* NOT REACHED */
#endif
}

fp_t
Scalb(fp_t x, fp_t y)
{
	y = Integer(y);		/* to match documentation in help_scalb() */
				/* and get consistency across platforms */
				/* NB: Use Integer(), not (LONG_LONG)(), to */
				/* handle NaN as documented */
#if defined(HAVE_SCALB)
	return (scalb(x,y));
#else
	return (x * pow(BASE, y));
#endif
}

fp_t
Significand(fp_t x)
{
	fp_t result;

#if defined(HAVE_SIGNIFICAND)
	result = significand(x);

	/* Some systems produce incorrect signs from significand(). */

	if ((result == FP(0.0)) && (copysign(FP(1.0), x) == FP(-1.0)))
	{
		result = FP(0.0);
		result = -result;	/* generate negative zero */
	}
#else
	/* Scale subnormal numbers into the normal range to avoid
	problems with subnormals in Trunc() and Log2() (e.g.,
	Compaq/DEC Alpha OSF/1 4.0). */

	if (IsNaN(x))
		result = copysign(FP(1.0),x) * NaN();
	else if (IsInf(x))
		result = copysign(FP(1.0),x) * Infinity();
	else if (x == FP(0.0))
		result = copysign(FP(1.0),x) * store(FP(0.0));
	else
	{
		FP_Parts d;
		fp_t xabs;
		int big_endian = IsBigEndian();

		xabs = (x < FP(0.0)) ? -x : x;

#if defined(HAVE_FP_T_DOUBLE)
		if (xabs < MinNormal())	/* scale to force normalization */
			x *= TWOTO52;
		d.v = x;
		d.i[1-big_endian] &= UL(0x000fffff);
		d.i[1-big_endian] |= (x < FP(0.0)) ? UL(0xbff00000) : UL(0x3ff00000);
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
		if (xabs < MinNormal())	/* scale to force normalization */
		{
			if (sizeof(d) == sizeof(d.v))	/* 128-bit */
				x *= TWOTO112;
			else				/* 80-bit Intel x86 format */
				x *= TWOTO64;
		}
		d.v = x;
		if (sizeof(d) == sizeof(d.v))	/* 128-bit */
		{
			d.i[3*(1-big_endian)] &= UL(0x0000ffff);
			d.i[3*(1-big_endian)] |= (x < FP(0.0)) ? UL(0xbfff0000) : UL(0x3fff0000);
		}
		else				/* 80-bit Intel x86 format */
		{
			d.i[2] &= UL(0xffff0000);
			d.i[2] |= (x < FP(0.0)) ? UL(0xbfff) : UL(0x3fff);
		}
#endif

#if defined(HAVE_FP_T_SINGLE)
		if (xabs < MinNormal())	/* scale to force normalization */
			x *= TWOTO23;
		d.v = x;
		d.i[0] &= UL(0x007fffff);
		d.i[0] |= (x < FP(0.0)) ? UL(0xbf800000) : UL(0x3f800000);
#endif

		result = d.v;
	}
#endif

	return (result);
}

fp_t
SNaN(void)
{
	fp_t maybe_NaN = NaN();
	static unsigned int SNaN_count = 0;

	if (IsNaN(maybe_NaN))		/* false on non IEEE 754 systems */
	{
		FP_Parts d;
		int big_endian = IsBigEndian();

		d.v = maybe_NaN;

		/* A NaN must have a nonzero significand, so we
		   choose to create up to 2^23 - 1 unique SNaNs.
		   For the moment, only two are generated by hoc,
		   for the constants SNAN and SNaN. */
		SNaN_count++;
		if (SNaN_count >= (int)0x7fffff) /* then 23-bit counter wrapped */
			SNaN_count = 1;

#if defined(HAVE_FP_T_DOUBLE)
		d.i[big_endian] = SNaN_count;
#if defined(__alpha)
		d.i[1 - big_endian] &= UL(~0x00080000);
#elif defined(__sparc)
		d.i[1 - big_endian] &= UL(~0x00080000);
#elif defined(__i860)
		d.i[1 - big_endian] &= UL(~0x00080000);
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		d.i[1 - big_endian] &= UL(~0x00080000);
#elif defined(__ia64)
		d.i[1 - big_endian] &= UL(~0x00080000);
#elif defined(__hppa)
		d.i[1 - big_endian] |=  0x00080000;
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v); /* No signaling NaN in the Intel IA-32 architecture */
#elif defined(__mips)
		(void)store(d.v);      /* No signaling NaN in the MIPS architecture */
#else  /* fall back to majority style, which could still be wrong */
		d.i[1 - big_endian] &= UL(~0x00080000);
#endif
#endif

#if defined(HAVE_FP_T_QUADRUPLE)
		d.i[3*big_endian] = SNaN_count;
#if defined(__alpha)
		d.i[3*(1 - big_endian)] &= UL(~0x00008000);
#elif defined(__sparc)
		d.i[3*(1 - big_endian)] &= UL(~0x00008000);
#elif defined(__i860)
		d.i[3*(1 - big_endian)] &= UL(~0x00008000);
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		d.i[3*(1 - big_endian)] &= UL(~0x00008000);
#elif defined(__ia64)
		d.i[3*(1 - big_endian)] &= UL(~0x00008000);
#elif defined(__hppa)
		d.i[3*(1 - big_endian)] |=  0x00008000;
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v); /* No signaling NaN in the Intel IA-32 architecture */
#elif defined(__mips)
		(void)store(d.v);      /* No signaling NaN in the MIPS architecture */
#else  /* fall back to majority style, which could still be wrong */
		d.i[3*(1 - big_endian)] &= UL(~0x00008000);
#endif
#endif

#if defined(HAVE_FP_T_SINGLE)
		d.i[0] = (d.i[0] & UL(0xff800000)) | (SNaN_count & UL(~0xff800000));
#if defined(__alpha)
		d.i[0] &= UL(~0x00400000);
#elif defined(__sparc)
		d.i[0] &= UL(~0x00400000);
#elif defined(__i860)
		d.i[0] &= UL(~0x00400000);
#elif defined(_POWER) || defined(__PPC) || defined(__powerpc)
		d.i[0] &= UL(~0x00400000);
#elif defined(__ia64)
		d.i[0] &= UL(~0x00400000);
#elif defined(__hppa)
		d.i[0] |=  0x00400000;
#elif defined(__i386) || defined(__i486) || defined(__i586) || defined(__i686)
		(void)store(d.v); /* No signaling NaN in the Intel IA-32 architecture */
#elif defined(__mips)
		(void)store(d.v);      /* No signaling NaN in the MIPS architecture */
#else  /* fall back to majority style, which could still be wrong */
		d.i[0] &= UL(~0x00400000);
#endif
#endif

		maybe_NaN = d.v;
	}
	return (maybe_NaN);
}

fp_t
Trunc(fp_t x)
{
	fp_t result;

#if defined(HAVE_TRUNC)
	result = trunc(x);
#else
	result = Integer(x);
#endif

	/* Some systems produce incorrect signs from trunc(). */

	if ((result == FP(0.0)) && (x < FP(0.0)))
	{
		result = FP(0.0);
		result = -result;	/* generate negative zero */
	}
	return (result);
}

fp_t
Unordered(fp_t x, fp_t y)
{
	return (IsNaN(x) || IsNaN(y));
}
